import mvpa.suite
import numpy as np
import os.path
import tabular as tb

import core.init
import tools.qFunctions as q

    
def run(
    rp,
    subROIs = False,
    suppressText = True,
    space = 'talairach',
    spm = False
    ):
    
    """
    Calculates mean coordinates and the number of voxels of each given ROI.
    
    **Parameters**
        rp: Namespace (required)
            Run parameters that are parsed from the command line
        subROIs: True or False
            If True, then subROIs are not combined together into an ROI
        suppressText: True or False
            If True, then nothing will be printed out
        space: talairach or native
            Choose the output to be either in native voxel space or in Talairach coordinates
        spm: True or False
            If True, then the coordinates in the voxel space are provided with
            indices +1 to match MatLab's convention of starting arrays from 1.
    
    """
    
    if subROIs: names = ['subjID','ROI','subROI','x','y','z','numVoxels']
    else: names = ['subjID','ROI','x','y','z','numVoxels']
    recs = []
    
    # allCoords = np.zeros((1,4))
    for subjIDno, subjID in enumerate(rp.subjID_list):
    
        for ROI_list in rp.rois:
            
            allROIs = []
            for thisROI in ROI_list[2]:
                allROIs.extend(q.listDir(core.init.paths['recDir'] %subjID,
                    pattern = thisROI + '\.nii', fullPath = True))
                    
            SForm = mvpa.suite.NiftiImage(allROIs[0]).getSForm()
            
            # check for overlap            
            mask = sum([mvpa.suite.NiftiImage(subROI).data for subROI in allROIs])            
            if not suppressText:
                overlap = mask > 2
                if np.sum(overlap) > 0:
                    print 'WARNING: Overlap in %(subjID)s %(ROI)s detected.'\
                    %{'subjID': subjID, 'ROI': ROI_list[1]}
                
            
            if not subROIs: allROIs = [mask]            
            for subROI in allROIs:
                
                if subROIs: subROIname = os.path.basename(os.path.abspath(subROI)).split('.')[0]
                else: subROIname = ROI_list[1]
                
                thisROI = mvpa.suite.NiftiImage(subROI)
                transROI = np.transpose(thisROI.data.nonzero())
                
                meanROI = np.mean(transROI,0)[1:]
                meanROI = meanROI[::-1] # reverse the order per convention
            
                # convert to the Talairach coordinates 
                if space == 'talairach':
                    meanROI = np.dot(SForm, np.concatenate((meanROI,[1]))) # convert
                    meanROI = meanROI[:-1] # remove the last coordinate (slice number)
                else:
                    meanROI = [m+spm for m in meanROI] # +1 to correct for SPM coords
                
                if subROIs:
                    recs.append((subjID,ROI_list[1],subROIname)+tuple(meanROI)+(transROI.shape[0],))
                else:
                    recs.append((subjID,subROIname)+tuple(meanROI)+(transROI.shape[0],))
    
    
        
    ROIparams = tb.tabarray(records = recs, names = names)
    
    if not suppressText:
        if subROIs: on = ['ROI','subROI']
        else: on = ['ROI']
        ROImean = ROIparams.aggregate(On = on, AggFunc = np.mean,
            AggFuncDict = {'subjID': lambda x: None})
        
        xyz = ROIparams[['x','y','z']].extract().reshape((len(rp.subjID_list),-1,3))
        xyzErr = np.std(xyz, axis = 0, ddof = 1)
        
        # sort ROImean
        numPerSubj = xyz.shape[1]
        order = ROIparams[:numPerSubj][on]
        order = order.addcols(range(len(order)), names=['order'])
        order.sort(order=on)
        ROImean.sort(order=on)
        ROImean = ROImean.addcols(order[['order']].extract(), names = 'order')
        ROImean.sort(order = 'order')
        
        lenROI = min([len(ROI) for ROI in ROImean['ROI']])
        if subROIs: lenSubROI = min([len(ROI) for ROI in ROImean['subROI']])
        print
        print ROIparams.dtype.names[1:]
        for i, line in enumerate(ROImean):
            print line['ROI'].ljust(lenROI+2),            
            if subROIs: print line['subROI'].ljust(lenSubROI+2),            
            print '%3d' %np.round(line['x']),
            print u'\xb1 %d  ' %np.round(xyzErr[i,0]),
            print '%3d' %np.round(line['y']),
            print u'\xb1 %d  ' %np.round(xyzErr[i,1]),
            print '%3d' %np.round(line['z']),
            print u'\xb1 %d  ' %np.round(xyzErr[i,2]),
            print '%4d' %np.round(line['numVoxels'])    
    
    return ROIparams