import numpy as np
import scipy.stats.stats as stats
from matplotlib import mpl
import matplotlib.pyplot as plt
import tabular as tb
import sys, os
sys.path.append('..')
import tools.qFunctions as q
import core.init


def plotrc(theme = 'default'):
    """
    Reads in the rc file
    Returns: plotOpt   
    """
    
    if theme == None or theme == 'default': themeFile = 'default'
    else: themeFile = theme
    
    rcFile = open(os.path.dirname(__file__) + '/' + themeFile + '.py')
    
    plotOpt = {}
    for line in rcFile.readlines():        
        line = line.strip(' \n\t')
        line = line.split('#')[0] # getting rid of comments
        if line != '': # skipping blank and commented lines
            
            linesplit = line.split(':',1)
            (key, value) = linesplit
            key = key.strip()
            value = value.strip()
            
            # try recognizing numbers, lists etc.
            try: value = eval(value)
            except: pass
            
            linesplit = key.rsplit('.',1)
            if len(linesplit) == 2:
                (plotOptKey, optKey) = linesplit
                if not plotOptKey in plotOpt: plotOpt[plotOptKey] = {}
                plotOpt[plotOptKey][optKey] = value
            elif key in ['cmap', 'simpleCmap']: plotOpt[key] = value
            else: print 'Option %s not recognized and will be ignored' %line
                        
    rcFile.close
    return plotOpt
    

def getColors(numColors, plotOpt, userOpt):
    """
    Returns either user specified colors (if the are enough of them) or
    generates them from a cmap
    """
    
    doCmap = False
    # if user has specified colors or a simpleCmap
    if 'colors' in userOpt:
        if type(userOpt['colors']) == np.ndarray:
            if len(userOpt['colors']) >= numColors:
                colors = userOpt['colors']
            else:
                print 'WARNING: Only %d ' %len(userOpt['colors']) +\
                    'were specified but there are %d ' %numColors +\
                    'categories. Default cmap colors will be used instead.'
                doCmap = True
        elif userOpt['colors'] in plotOpt['simpleCmap'].keys():
            colors = plotOpt['simpleCmap'][userOpt['colors']]
        else:
            print 'WARNING: User defined colors were not recognized. '+\
                'Default colors will be used instead.'
            doCmap = True
    else: doCmap = True
    
    # if colors were not specified by user
    if doCmap:
        thisCmap = plotOpt['cmap']
        if type(thisCmap).__name__ in ('list','tuple'):
            cmap = cycle(thisCmap)
            colors = np.array([cmap.next for i in range(numCat)])
        else:
            cmap = mpl.cm.get_cmap(thisCmap)        
            if cmap == None:
                print 'Colormap %s does not exist' %value
                print "Default colormap ('Paired') will be used instead"
                cmap = mpl.cm.get_cmap('Paired')            
            norm = mpl.colors.Normalize(0, 1)
            z = np.linspace(0, 1, numColors + 2)
            z = z[1:-1]
            colors = cmap(norm(z))
            
    return colors
    
    
def prettyPlot(
    x,
    y = None,    
    yerr = None,
    plotType = 'errorbar',
    fig = None,
    subplotno = 111,
    max_y = None,
    theme = 'default',
    userOpt = None
    ):
    
    """
    Automatically makes beautiful plots and with error bars.
    
    **Parameters**
    
        x: numpy.ndarray
            Either x or y values
        y: **None** or numpy.ndarray
            + if **None**, then x value is assumed to be y, while x is generated automatically    
            + else it is simpy a y value
        yerr: **None** a 2D numpy.ndarray or list
            Specifies y error bars. If **None**, then no error bars are plotted
        plotType: 'errorbar' or 'bar'
            + 'errorbar' plots lines with error bars
            + 'bar' plots bars with error bars
        subplotno: int
            Use 111 for a single plot; otherwise specify subplot position
        max_y: **None** or int
            Use to indicate the maximum allowed ytick value.
            Useful when plotting accuracy since yticks should not go above 1
        theme: str
            Specifies the file name where all the custom plotting parameters are stores
        userOpt: dict
            Any plotting options that should override default options
        
    """
    
    if y == None: # meaning: x not specified
        y = x
        if y.ndim == 1:
            x = np.arange(y.shape[0])+1
            numCat = 1
        else:
            x = np.arange(y.shape[1])+1
            numCat = y.shape[0]
        # this is because if y.ndim == 1, y.shape = (numb,) instead of (1,num)
        
    numXTicks = x.shape[0]
    
    plotOpt = plotrc(theme)
    
    if not 'cmap' in plotOpt: plotOpt['cmap'] = 'Paired'    
    if not 'hatch' in plotOpt: plotOpt['hatch'] = ['']*numCat
        
    if not 'subplot' in plotOpt: plotOpt['subplot'] = {}
    if not 'xticks' in plotOpt['subplot']:
        if numXTicks > 1:
            plotOpt['subplot']['xticks'] = x
        else:
            plotOpt['subplot']['xticklabels'] = ['']
    if not 'legend' in plotOpt: plotOpt['legend'] = ['']    
    
    # x limits        
    xMin = x.min()
    xMax = x.max()
    if xMax == xMin: plotOpt['subplot']['xlim'] = (-.5,2.5)
    elif plotType == 'errorbar':
        plotOpt['subplot']['xlim'] = (xMin - (xMax - xMin)/(2.*(len(x)-1)),
            xMax + (xMax - xMin)/(2.*(len(x)-1)))
    elif plotType == 'bar':
        plotOpt['subplot']['xlim'] = (xMin - (xMax - xMin)/(1.5*(len(x)-1)),
            xMax + (xMax - xMin)/(1.5*(len(x)-1)))
    
    # y limits
    if yerr == None or not np.isfinite(yerr.max()):
        yMin = y.min()
        yMax = y.max()
    else:
        yMin = (y - yerr).min()
        yMax = (y + yerr).max()
    if yMax == yMin: plotOpt['subplot']['ylim'] = (0, yMax + yMax/4.)
    else: plotOpt['subplot']['ylim'] = (yMin - (yMax - yMin)/4., yMax + (yMax - yMin)/2.)
        
    
    # overwrite plotOpt by userOpt
    if userOpt != None:
        for (key, value) in userOpt.items():
            if type(value)==dict:
                if not key in plotOpt: plotOpt[key] = {}
                for (key2, value2) in value.items():
                    plotOpt[key][key2] = value2
            else: plotOpt[key] = value
    
    
    # set all values recognized by mpl
    # this is stupid but not all values are recognized by mpl
    # thus, we later set the unrecognized ones manually
    for key, value in plotOpt.items():
        if not key in ['cmap', 'colors', 'hatch', 'legend.names', 'other']:            
            for k, v in value.items():            
                fullKey = key + '.' + k            
                if fullKey in mpl.rcParams.keys():
                    try: mpl.rcParams[fullKey] = v
                    except: mpl.rcParams[fullKey] = str(v) # one more stupidity in mpl
                        # for ytick, a string has to be passed, even if it is a number
    plotOpt['colors'] = getColors(numCat, plotOpt, userOpt)                    
    
    
    # make a new figure
    if fig == None:
        fig = plt.figure()
        fig.canvas.set_window_title('Results')    
    ax = fig.add_subplot(subplotno)
    
    output = []    
    
    # generate plots
    if plotType == 'errorbar':
        
        if yerr != None:
            for i in range(numCat):
                output.append(ax.errorbar(
                    x,
                    y[i,:],
                    yerr = yerr[i,:],
                    color = plotOpt['colors'][i],
                    markerfacecolor = plotOpt['colors'][i],                    
                    **plotOpt[plotType]
                    )
                )
        else:        
            for i in range(numCat):                
                output.append(ax.plot(
                    x,
                    y[i,:],
                    color = plotOpt['colors'][i],
                    markerfacecolor = plotOpt['colors'][i]
                    )
                )
                
    elif plotType == 'bar':
    
        barWidth = (1-.2)/numCat       # the width of the bars
        middle = x - numCat*barWidth/2.
        
        for i in range(numCat):            
            output.append(ax.bar(
                middle+i*barWidth,
                y[i,:],
                barWidth,
                facecolor = plotOpt['colors'][i],
                yerr = yerr[i,:],
                hatch = plotOpt['hatch'][i],
                **plotOpt[plotType]
                )
            )
            
    
    # set up legend
    if plotOpt['subplot'].get('legend.names'):
        if len(plotOpt['subplot']['legend.names']) > 1:
            outLeg = [i[0] for i in output]
            leg = ax.legend(outLeg, plotOpt['subplot']['legend.names'])
            leg.draw_frame(False)
        del plotOpt['subplot']['legend.names']
        
    # draw other things
    if 'other' in plotOpt['subplot']:
        for item in q.listify(plotOpt['subplot']['other']): eval(item)
        del plotOpt['subplot']['other']
        
    # set the remaining subplot options    
    ax.set(**plotOpt['subplot'])
    
    if not 'yticks' in plotOpt['subplot'] and not 'yticklabels' in plotOpt['subplot']:
        import decimal
        yticks = ax.get_yticks()
        if max_y != None: yticks = yticks[yticks <= max_y]
        else: yticks = yticks[yticks <= yticks[-2]]
        # now some fancy way to make pretty labels
        kk = min([decimal.Decimal(str(t)).adjusted() for t in yticks])        
        lens2 = np.array([t for t in yticks if len(('%g' %(t/10**kk)).split('.')) == 1])
        spacing = len(lens2)/6 + 1
        if max_y != None: use = lens2[np.arange(len(lens2)-1,-1,-spacing)]
        else: use = lens2[np.arange(0,len(lens2),spacing)]
        
        ax.set_yticks(use)
    
    return fig

    
def calcInd(
    cleanData,
    plotData,
    xAxis,
    categories,
    values,
    xAxisOrder = None,
    catOrder = None
    ):
    """
    
    """
    
    if cleanData[values].dtype.type == np.dtype(str): aggFun = lambda x: float(len(x))
    else: aggFun = np.mean
    
    # Count the number of ALL instances to get the percentage
    if cleanData[values].dtype.type == np.dtype(str):
        countAgg = plotData.aggregate(On = xAxis+categories,
            AggFuncDict = {values: lambda x: float(len(x))})
    
    # This is where the main agggregation of data happens
    agg = cleanData.aggregate(On = xAxis+categories,
        AggFuncDict = {values: aggFun})
    
    # Finding proportion correct
    if cleanData[values].dtype.type == np.dtype(str):
        agg[values] = agg[values] / countAgg[values]
    
    # Here we calculate the average across subjects
    #plotData = agg.aggregate(On = xAxis+categories, AggFuncDict = {values: np.mean})
    
    sortOrder = []
    if xAxisOrder != None:
        ord = []        
        for cat in agg[xAxis]:
            for i, item in enumerate(xAxisOrder):
                if str(cat[0]).lower() == str(item[0]).lower():
                    ord.append(i)
                    break
        if len(agg) != len(ord):
            print 'ERROR: Your specified %s labels do not match the ones extracted from the input file' %xAxis[0]
            print [c[0] for c in agg[xAxis]]
            print [c[1] for c in xAxisOrder]
            sys.exit()
        agg = agg.addcols(ord,names=['xAxisOrderINNERVAR'])
        sortOrder.append('xAxisOrderINNERVAR')
    else: sortOrder.extend(xAxis)
    
    
    if catOrder != None:
        ord = []        
        for cat in agg[categories]:
            for i, item in enumerate(catOrder):            
                if cat[0].lower() == item[0].lower():
                    ord.append(i)
                    break
        if len(agg) != len(ord):
            print 'ERROR: Your specified %s labels do not match the ones extracted from the input file' %categories[0]
            print [c[0] for c in agg[categories]]
            print [c[1] for c in catOrder]
            sys.exit()
        agg = agg.addcols(ord,names=['catOrderINNERVAR'])
        sortOrder.append('catOrderINNERVAR')
    else: sortOrder.extend(categories)
    
    agg.sort(order = sortOrder)
    agg = agg[[values]].extract()
    # Each series goes into a column
    #numXTickLabels = cleanData.aggregate(On = xAxis)[xAxis].extract().size
    #numCatLabels = cleanData.aggregate(On = categories)[categories].extract().size
    numXTickLabels = cleanData.aggregate(On = xAxis)[xAxis].extract().size
    numCatLabels = cleanData.aggregate(On = categories)[categories].extract().size
    agg = agg.reshape(numXTickLabels,numCatLabels).transpose()
    
    return agg
    
    
def plotTabular(
    plotData,
    fig = None,
    subplotno = 111,
    filter = None,
    xAxis = None,
    categories = None,
    values = None,
    valueFilter = None,
    max_y = None,
    yerrName = None,
    plotType = 'errorbar',
    theme = 'default',
    textOnly = False,
    savePlot = None,
    userOpt = {}
    ):
    
    """
    **Parameters**
    
        plotData: tabarray (required)
            Data that is to be plotted
        textOnly: True or False
            + If True, then plots are shown and values printed in the command line
            + If False, then only values are printed in the command line
        userOpt: dict
        
    """
    
    def ttestPrint(title = '****', values = None, plotOpt = None, type = None):
    
        # check if there are any negative t values (for formatting purposes)
        if np.any([np.any(val < 0) for val in values]): neg = True            
        else: neg = False
            
        print '\n' + title
        for xi, xticklabel in enumerate(plotOpt['subplot']['xticklabels']):
            print xticklabel
            
            maxleg = max([len(leg) for leg in plotOpt['subplot']['legend.names']])
            if type == 1: legendnames = ['%*s' %(maxleg,p) for p in plotOpt['subplot']['legend.names']]
            elif type == 2:
                pairs = q.combinations(plotOpt['subplot']['legend.names'],2)                
                legendnames = ['%*s' %(maxleg,p[0]) + ' vs ' + '%*s' %(maxleg,p[1]) for p in pairs]                
                
            for yi, legendname in enumerate(legendnames):
                if values[1][yi,xi] < .001: star = '***'
                elif values[1][yi,xi] < .01: star = '**'
                elif values[1][yi,xi] < .05: star = '*'
                else: star = ''
                
                if neg and values[0][yi,xi] > 0:
                    outputStr = '    %(s)s: t(%(d)d) =  %(t).3f, p = %(p).3f %(star)s'
                else:
                    outputStr = '    %(s)s: t(%(d)d) = %(t).3f, p = %(p).3f %(star)s'
                    
                print outputStr \
                    %{'s': legendname, 'd':(d-1), 't': values[0][yi,xi],
                    'p': values[1][yi,xi], 'star': star}
    
    
    # work out the order of xAxis
    xAxisOrder = None
    # if xAxisOrder is specified explicity as a part of xAxis
    if type(xAxis) == list and len(xAxis) > 1: 
        xAxisOrder = xAxis[1]
        xAxis = xAxis[0]
    xAxis = q.listify(xAxis)
    
    # make xAxisOrder contain original label and renaming
    if xAxisOrder != None:
        xAxisOrder = q.listify(xAxisOrder,level=1)
        tempXAxisOrder = []
        for thisItem in xAxisOrder:
            if len(thisItem) == 1: # renaming not specified
                tempXAxisOrder.append([thisItem[0], thisItem[0]])
            else:
                tempXAxisOrder.append(thisItem)
        xAxisOrder = tempXAxisOrder
        
        # make a filter only from xAxis labels present in the xAxisOrder
        first = True
        for item in xAxisOrder:
            if first:
                filter2 = plotData[xAxis[0]] == item[0]
                first = False
            else: filter2 = np.logical_or(filter2,plotData[xAxis[0]] == item[0])
        plotData = plotData[filter2]
    
    
    # work out the order of categories
    catOrder = None
    # if catOrder are specified explicity as a part of categories
    if type(categories) == list and len(categories) > 1:
        catOrder = categories[1]
        categories = categories[0]
    categories = q.listify(categories)
    
    
    # make categories contain original label and renaming
    if catOrder != None:
        catOrder = q.listify(catOrder,level=1)
        tempCatOrder = []
        for thisItem in catOrder:
            if len(thisItem) == 1: tempCatOrder.append([thisItem[0], thisItem[0]]) # no renaming
            else: tempCatOrder.append(thisItem)
        catOrder = tempCatOrder        
        catLabels = np.array([lab[1] for lab in catOrder])
        
        # make a filter only from category labels present in the catOrder
        first = True
        for item in catOrder:
            if first:
                filter3 = plotData[categories[0]] == item[0]
                first = False
            else: filter3 = np.logical_or(filter3, plotData[categories[0]] == item[0])
        plotData = plotData[filter3]
        
    
    # Apply filters to select relevant columns from plotData
    if filter != None: plotData = plotData[eval(filter)]
    
    
    # if filterXAxis:
        # # make a filter only from xAxis labels present in the xAxisOrder
        # first = True
        # for item in xAxisOrder:
            # if first:
                # filter2 = allData[xAxis[0]] == item[1]
                # first = False
            # else: filter2 = np.logical_or(filter2,allData[xAxis[0]] == item[1])
    # else: filter2 = True
    
    # if filterCat:
        # # make a filter only from category labels present in the catOrder
        # first = True
        # for item in catOrder:
            # if first:
                # filter3 = allData[categories[0]] == item[1]
                # first = False
            # else: filter3 = np.logical_or(filter3, allData[categories[0]] == item[1])
    # else: filter3 = True
    
    # if filterXAxis or filterCat:
        # allData = allData[np.logical_and(filter2,filter3)]
        
    # Apply extra filters to choose which conditions will be plotted
    # note that (some of the) values will still be calculated using plotData
    if valueFilter != None: cleanData = plotData[eval(valueFilter)]
    else: cleanData = plotData
    
    
    # set x-axis labels
    hasXTickLabels = False
    if 'subplot' in userOpt:
        if 'xticklabels' in userOpt['subplot']:
            hasXTickLabels = True
            
    if hasXTickLabels:
        xTickLabels = userOpt['subplot']['xticklabels']    
    elif xAxisOrder == None: # order not specified
        xTickLabels = cleanData.aggregate(On = xAxis)[xAxis].extract()
        if xTickLabels.size < 2: xTickLabels = np.array([''])
        xTickLabels = np.array([str(tc) for tc in xTickLabels])
    else:
        xTickLabels = np.array([lab[1] for lab in xAxisOrder])
            
            
    # categories order not specified, get labels from the data        
    if catOrder == None: 
        catLabels = cleanData.aggregate(On = categories)[categories].extract()
        if catLabels.size == 0: catLabels = np.array([''])
        elif catLabels.size == 1: catLabels = catLabels.reshape((1))
        
    
    plotOpt = {}    
    plotOpt['subplot'] = {}    
    plotOpt['subplot']['legend.names'] = map(str.lower,catLabels)
    if xTickLabels.size < 2: plotOpt['subplot']['xlabel'] = ''
    else: plotOpt['subplot']['xlabel'] = xAxis[0]
    plotOpt['subplot']['xticks'] = np.arange(xTickLabels.size)+1
    plotOpt['subplot']['xticklabels'] = xTickLabels#np.hstack((np.array([None]),xTickLabels))
    plotOpt['subplot']['title'] = values
    
    
    for (key, value) in userOpt.items():
        if key == 'subplot':
            for (okey, ovalue) in userOpt[key].items():
                plotOpt[key][okey] = ovalue
        else: plotOpt[key] = value
    
    # Set up error bar information
    yerrList = q.listify(cleanData.aggregate(On = yerrName)[[yerrName]].extract().tolist())
    numYerr = len(yerrList)
    
    agg = np.zeros((numYerr,catLabels.size,xTickLabels.size))
         
    for i, ind in enumerate(yerrList):
        agg[i,:,:] = calcInd(cleanData[cleanData[yerrName] == ind],\
            plotData[plotData[yerrName] == ind], xAxis,categories, values, 
             xAxisOrder, catOrder)
    
    plotData = np.sum(np.array(agg),0)/numYerr
    
    
    # # Here is the standard error across subjID calculated
    # plotDataYerr = agg[[values]].extract()
    # #import pdb; pdb.set_trace()
    
    # plotDataYerr = plotDataYerr.reshape(numYerr,xTickLabels.size,catLabels.size)
    # plotDataYerr = np.std(plotDataYerr, axis = 0)/np.sqrt(plotDataYerr.shape[0]-1)
    # # Each series goes into a column
    # plotDataYerr = plotDataYerr.transpose()
    
    # Here is the standard error of the mean across yerr calculated
    plotDataYerr = np.std(agg, axis = 0, ddof = 1)/np.sqrt(agg.shape[0])
    
    doTTest = True
    if doTTest:
        d = agg.shape[0]
        
        # 2-tail T-Test
        ttest = (np.zeros((agg.shape[1]*(agg.shape[1]-1)/2, agg.shape[2])),
                 np.zeros((agg.shape[1]*(agg.shape[1]-1)/2, agg.shape[2])))
        ii = 0
        for c1 in range(agg.shape[1]):
            for c2 in range(c1+1,agg.shape[1]):
                thisTtest = stats.ttest_rel(agg[:,c1,:], agg[:,c2,:], axis = 0)
                ttest[0][ii,:] = thisTtest[0]
                ttest[1][ii,:] = thisTtest[1]
                ii += 1
        ttestPrint(title = '**** 2-tail T-Test of related samples ****',
            values = ttest, plotOpt = plotOpt,
            type = 2)
        
        # One-sample t-test
        m = .5
        oneSample = stats.ttest_1samp(agg, m, axis = 0)
        ttestPrint(title = '**** One-sample t-test: difference from %.2f ****' %m,
            values = oneSample, plotOpt = plotOpt, type = 1)
            
    
    if max_y == None and cleanData[values].dtype.type == np.dtype(str): max_y = 1
    
    
    fig = prettyPlot(plotData, plotType = plotType, yerr = plotDataYerr,
        fig = fig, subplotno = subplotno, max_y = max_y, userOpt = plotOpt)
    
    print ''
    print 'Results:'
    print np.transpose(plotData)
    if savePlot != None: plt.savefig(savePlot, bbox_inches = 'tight')
    if not textOnly: plt.show()
    
    return fig
    