import mvpa.suite
import numpy as np
import matplotlib.pyplot as plt
import tabular as tb
import random
import csv
import os, sys

import core
import dataAnalysis as da
import tools.qFunctions as q
import ROIparams
import behav

        
def makeNim(
    subjID = None,
    runType = None,
    tabData = None,
    ROIs = None,
    filter = None,
    shiftTp = 2
    ):
    """
    Produces a detrended dataset with info for classifiers.
    
    **Parameters**
        subjID: str
        runType: str
        tabData: tabular.tabarray
            A tabarray with all information about conditions.
            Must have 'sessNo', 'cond', and 'accuracy' columns
        ROIs: list
            A pattern of ROI file patterns to be combined into one ROI
        filter: **None** or str (a rule)
            Filter out particular datapoints
        shiftTp: int
            Specifies how many TRs after the onset of a stimulus to extract the signal
    
    **Returns**
        nim: NiftiDataset
        
    """
    
    # make a mask by combining all ROIs
    allROIs = []
    for ROI in ROIs:
        allROIs.extend(q.listDir(core.init.paths['recDir'] %subjID,
        pattern=ROI+'\.nii', fullPath = True))
    
    thisROI = sum([mvpa.suite.NiftiImage(roi).data for roi in allROIs])
    thisMask = mvpa.suite.NiftiImage(thisROI)    
    # change 4D mask to 3D by removing the first dim (time)
    maskData = thisMask.data[:,:,:,::-1] # flip ROI because functional data is flipped
    thisMask.setDataArray(maskData.reshape(maskData.shape[1:]))
    
    
    # find all functional runs of a given runType
    allImg = q.listDir(core.init.paths['dataFmriDir'] %subjID, pattern = 'swra.*' + \
        runType + '\.img', fullPath = True)        
    
    # Load in data for all runs and all ROIs 
    chunkCount = 0
    first = True
    for thisImg in allImg:
       
        thisData = tabData[tabData['sessNo'] == int(thisImg.split('_')[-2])]
        numRec = len(thisData)
        # shift data by shiftTp to compensate for the hemodynamic lag
        thisData = thisData[numRec-shiftTp: numRec].addrecords(\
            thisData[0:numRec-shiftTp])
                 
        # load the appropriate func file with a mask
        tempNim = mvpa.suite.NiftiDataset(            
            samples = thisImg,
            labels = thisData['cond'],
            chunks = chunkCount,
            mask = thisMask
            )
        
        # get scaled and shifted data
        tempNim.samples = tempNim.niftihdr['scl_inter'] + \
            tempNim.samples.astype('float') * tempNim.niftihdr['scl_slope']
        # NOTE: getScaledData() would be a better method to achive this
        # but it is only available for NiftiImage
        
        # choose samples only based on certain conditions        
        if filter != None: tempNim = tempNim.selectSamples(eval(filter))
        
        # combine all functional runs into one massive NIfTI Dataset
        if first:
            nim = tempNim
            first = False
        else: nim += tempNim
        
        chunkCount += 1
    
    
    # detrend per run with a second order polynomial
    polyord = 2
    # perform detrending
    dtr = mvpa.suite.detrend(nim, perchunk = True, polyord = polyord)    
    # assign correct detrented value to the nim samples
    dtrValue = dtr[0][0].reshape( chunkCount, polyord+1, dtr[0][0].shape[1] )[:,0,:]
    for i, chunk in enumerate(nim.UC):
        selSamples = np.array([chunk == c for c in nim.C], dtype='bool')
        nim.samples[selSamples,:] += dtrValue[i,:]    
    
    return nim
    

def makeNim3(
    subjID = None,
    runType = None,
    ROIs = None,
    tabData = None,
    filter = None,
    shiftTp = 2
    ):
    """
    Produces a detrended dataset with info for classifiers.
    
    **Parameters**
        rp: Namespace (required)
            Run parameters that are parsed from the command line
        ROIs: str
        tabData: **None** or tabular.tabarray
            A tabarray with all information about conditions.
            Must have 'sessNo', 'cond', and 'accuracy' columns.
            If **None**, then read in automatically
        filter: **None** or str (a rule)
            Filter out particular datapoints
        shiftTp: int
            Specifies how many TRs after the onset of a stimulus to extract the signal
    
    **Returns**
        nim: NiftiDataset
        
    """
    
    if tabData == None: tabData = behav.getBehavData(rp)
    
    # make a mask by combining all ROIs
    allROIs = []
    for ROI in ROIs:
        allROIs.extend(q.listDir(core.init.paths['recDir'] %subjID, pattern=ROI+'\.nii', fullPath = True))
    
    #thisROI = np.array(sum([mvpa.suite.NiftiImage(roi).data for roi in allROIs]))    
    #thisMask = mvpa.suite.NiftiImage(thisROI)
    #plt.imshow(np.mean(thisMask.data[0],0),cmap='gray',origin='lower')
    #plt.show()
    nim_list = []
    for thisROI in allROIs:
        
        thisMask = mvpa.suite.NiftiImage(thisROI)
        
        # change 4D mask to 3D by removing the first dim (time)
        maskData = thisMask.data[:,:,:,::-1] # flip ROI because functional data is flipped
        thisMask.setDataArray(maskData.reshape(maskData.shape[1:]))
        
        
        # find all functional runs of a given runType
        allImg = q.listDir(core.init.paths['dataFmriDir'] %subjID, pattern = 'swra.*' + \
            runType + '\.img', fullPath = True)
            
        
        # Load in data for all runs and all ROIs 
        chunkCount = 0
        first = True
        for thisImg in allImg:
           
            thisData = tabData[tabData['sessNo'] == int(thisImg.split('_')[-2])]
            numRec = len(thisData)
            # shift data by shiftTp to compensate for the hemodynamic lag
            thisData = thisData[numRec-shiftTp: numRec].addrecords(\
                thisData[0:numRec-shiftTp])
                
            # For the event-related designs. Currently too slow to be used
            # events = [mvpa.suite.Event(
                # onset       = tabData['onset'][i],
                # duration    = tabData['dur'][i],
                # label       = tabData['cond'][i],
                # chunk       = chunkCount)
                # for i in range(len(tabData))]
                     
            # load the appropriate func file with a mask
            tempNim = mvpa.suite.NiftiDataset(            
                samples = thisImg,
                labels = thisData['cond'],
                chunks = chunkCount,
                mask = thisMask
                )
            
            
            #if thisROI == '..\\data\\confSup_02\\reconstruction/lh_V3d.nii':
            #    print np.sum(tempNim.samples!=0)
            #    import pdb; pdb.set_trace()
            
            # get scaled and shifted data
            tempNim.samples = tempNim.niftihdr['scl_inter'] + \
                tempNim.samples.astype('float') * tempNim.niftihdr['scl_slope']
            # NOTE: getScaledData() would be a better method to achieve this
            # but it is only available for NiftiImage
            
            
            # left-right flip along the x-axis
            #rev = tempNim.mapReverse(tempNim.samples)
            #tempNim.samples = tempNim.mapForward(rev[:,:,:,::-1])
            
            #plt.figure()
            #print thisImg
            #thisNim = mvpa.suite.NiftiImage(thisImg)
            #plt.imshow(np.mean(tempNim.mapReverse(tempNim.samples)[0],0),cmap='gray',origin='lower')
            #plt.show()
            
            # plt.figure()
            # print thisMask.filename            
            # plt.imshow(np.mean(thisMask.data,0),cmap='gray',origin='lower')
            #plt.show()
            #import pdb; pdb.set_trace()
            
            #origNim = mvpa.suite.samples_original(nim)
            #print np.sort(tempNim.samples[0,:])[:10]
            #    
            # choose samples only based on certain conditions        
            if filter != None:
                tempNim = tempNim.selectSamples(eval(filter))
            
            # combine all functional runs into one massive NIfTI Dataset
            if first:
                nim = tempNim
                first = False
            else: nim += tempNim
            
            chunkCount += 1
        
        
        
        # detrend per run with a second order polynomial
        polyord = 2
        # perform detrending
        dtr = mvpa.suite.detrend(nim, perchunk = True, polyord = polyord)    
        # assign correct detrented value to the nim samples
        dtrValue = dtr[0][0].reshape( chunkCount, polyord+1, dtr[0][0].shape[1] )[:,0,:]
        for i, chunk in enumerate(nim.UC):
            selSamples = np.array([chunk == c for c in nim.C], dtype='bool')
            nim.samples[selSamples,:] += dtrValue[i,:]    
        nim_list.append(nim)
        
    return nim_list

    
    
def wrapperClf(smp, pairs, splits, clfList, selRange, recPre = []):
    
    out = []
    for trainPair, testPair in pairs:
    
        nimSel = []
        inds = {}
        for count, lab in enumerate(trainPair+testPair):
        
            if not lab in inds:
                # sample randomly time points
                ind = np.random.permutation(np.arange(len(smp[lab])))
                ind = ind[:selRange]            
                thisSel = smp[lab][ind,:]
                
                # normalization / z-scoring
                for tp in range(len(thisSel)):
                    thisSel[tp,:] = (thisSel[tp,:] - np.mean(thisSel[tp,:])) / \
                                     np.std(thisSel[tp,:],ddof=1)
                                     
                nimSel.append(thisSel)
                inds[lab] = count
                
            else: nimSel.append(nimSel[inds[lab]])
        
        temp = []        
        for allTp, invAllTp in splits:
            training = mvpa.suite.Dataset(
                samples = np.concatenate((nimSel[0][allTp],nimSel[1][allTp])),
                labels = [trainPair[0]]*sum(allTp) + [trainPair[1]]*sum(allTp))
            validation = mvpa.suite.Dataset(
                samples = [np.mean(nimSel[2][invAllTp],axis=0), \
                    np.mean(nimSel[3][invAllTp],axis=0)],
                labels = trainPair)    
            
            for c, thisClf in enumerate(clfList):
                thisClf.train(training)
                prediction = np.mean(thisClf.predict(validation.samples) == validation.labels)
                if len(temp)-1 < c: temp.append([prediction])
                else: temp[c].append(prediction)
        
        meanPerf = np.mean(np.array(temp), axis = 1)
        out.append(tuple(recPre + list(trainPair) + list(testPair) + meanPerf.tolist()))

    return out


def run(
    rp,
    allConds = q.OrderedDict([
                        ('parts', [[1,2,3,4],[1,2,3,4]]),
                        ('whole', [[5,6,7,8],[5,6,7,8]])
                     ]),
    clf = 'LinearNuSVMC',
    repeat = 10,
    splitRule = (10,10),
    filter = None,
    acc = None,
    plot = True,
    numVoxels = 'all',
    condsForRef = 'sel'
    ):
    
    """
    Performs an SVM classification.
    
    **Parameters**
    
        rp: Namespace (required)
            Run parameters that are parsed from the command line
        allConds: OrderedDict
            An ordered dictionary containing the lists of conditions that will go for
            training and testing
        clf: 'SMLR', 'LinearCSVMC', 'LinearNuSVMC', 'RbfNuSVMC', or 'RbfCSVMC', or a list of them
            A list of classifiers to be used
        repeat: int
            how many times repeat SVM training+testing cycle (with a different selection of timepoints)
        splitRule: list or 'min'
            + a list of number of chunks (~runs) and size of each chunk
            + or 'min' - take the least number across all conditions, split by sqrt
        filter: **None** or a rule
            Specify which data should be taken.
            + **None** means all data will be used.
            + Rule: e.g., "thisData['accuracy'] == 'Correct'" (only correct responses taken)
        acc: **None**, 'max', or float
            + **None** means that all data will be used
            + 'max': take max accuracy, typically limited by the whole condition
            + float: e.g., .95 (take 95% of correct responses and 5% of incorrect)
        plot: **True** or **False**
            Whether to plot the output or not
        numVoxels: 'all', 'min' or int
            Specify 
            + 'all': use all voxels available
            + 'min': use the minimal number of voxels across all ROIs for all ROIs
            + int: use that many voxels for all ROIs
        condsForRef: 'all', 'sel', or 'min'
            When getting accuracy or selecting voxels for analysis, specify whether 
            these analyses should depend on all conditions ('all') or only on the 
            ones specified at allConds ('sel')
            
    """
    
    if rp.i != None: # simply plot from the existing analysis file
        plotData = tb.tabarray(SVfile = rp.i)
    else:
        # send it all to the classifier
        names = ['subjID','ROI','ROIfiles','numVoxels','type','session',\
            'train1', 'train2', 'test1', 'test2']        
        clf = q.listify(clf)
        names += clf    
        if rp.numVoxels != None:
            try: numVoxels = int(rp.numVoxels) #if specified as an int
            except: numVoxels = rp.numVoxels
        
        tabData = behav.getBehavData(rp)
        #import pdb; pdb.set_trace()
        params = ROIparams.run(rp, space = 'native')
        params.sort(order=['subjID','numVoxels'])
        
        # have more iterations within an ROI if random voxels sampling is specified
        if numVoxels == 'minRandom':
            numRoiIter = int(np.sqrt(repeat))
            repeat /= numRoiIter
            print 'INFO: Repetions over ROIs changed to %d ' %numRoiIter +\
            'and repetions within ROIs changed to %d' %repeat
        else: numRoiIter = 1
        
        # read in scan data and perform SVMs for each subject each ROI group    
        dataOut = []
        for subjIDno, subjID in enumerate(rp.subjID_list):
            print ''
            print subjID            
            
            # for each condition, count how many Correct, Incorrect and No response there are
            # but remove condition 0 (fixation) and whatever is specified by filter
            thisFilter = (tabData['subjID'] == subjID) & (tabData['cond'] != 0)
            if filter != None: thisFilter = thisFilter & filter
            thisData = tabData[thisFilter]
            
            cac = thisData[['subjID','cond','accuracy']].aggregate(
                On = ['accuracy','cond'],
                AggFunc = lambda x: len(x)
                )
            cac.renamecol('subjID','count')
            
            # specify how (train, test) pairs will be made
            if splitRule == 'min':
                # take the least number of timepoints available across all conditions
                # you can therefore have an equal number of timepoints for all conditions
                # and that number is maximal
                # this is useful, for example, if you're only looking at correct or incorrect trials
                
                # find the number of each condition occurrence, and get the min
                minCount = np.min(cac['count'])
                
                # produce a reasonable splitting rule
                r = int(np.sqrt(minCount))
                split = (r, minCount/r)
                print 'INFO: Using the split ' + str(split)                                
                
            else: split = splitRule
            
            
            selRange = split[0]*split[1] # how many timepoints to select
            clfList = [eval('mvpa.suite.' + thisClf + '()') for thisClf in clf]
            
            
            # prepare indices for training and testing sets
            # notice that these indeces specify only the splits
            # and not the actual indices of the timepoints to choose
            splits = []    
            for chunk in range(split[0]):
            
                # training set indices
                allTp = np.ones((selRange,), dtype='Bool')
                allTp[split[1]*chunk:split[1]*(chunk+1)] = np.zeros((split[1],))
                # testing set indices
                invAllTp = (allTp+1)%2
                invAllTp = invAllTp.astype(bool)        
                splits.append([allTp, invAllTp]) # this is [trainSet, testSet]
                
            
            thisROIparams = params[params['subjID']==subjID]
                        
            # go through each ROI and perform SVM
            print ''
            print 'Running an SVM on...'
            for roiIter in range(numRoiIter):
                # specify which voxels should be used
                
                # some specified number of voxels
                if type(numVoxels) == int: useVoxels = numVoxels
                # use the most voxels possible, select random
                elif numVoxels == 'minRandom':
                    useVoxels = thisROIparams['numVoxels'][0] # note: it is already sorted
                    print 'INFO: Number of voxels used is %d' %useVoxels
                # use all voxels
                elif numVoxels == 'all': useVoxels = -1 ## THIS IS WRONG!!!!
                else:
                    print 'ERROR: Cannot process the number of voxels specified'
                    sys.exit()
                    
                selVoxels = {}
                for thisROI, thisSize in thisROIparams[['ROI','numVoxels']]:
                    if thisSize < useVoxels:
                        print 'ERROR: Number of voxels to use exceed the available number of voxels %d' %thisSize
                        sys.exit()
                    selVoxels[thisROI] = np.random.permutation(np.arange(thisSize))[:useVoxels]
                    
                for ROI_list in rp.rois:
                    print ROI_list[1],
                    print ':',
                    thisNim = makeNim(subjID = subjID, runType = rp.runType, tabData = tabData[tabData['subjID']==subjID],
                        ROIs = ROI_list[2], filter = filter)
                        
                    # generate a dict = {cond: [2D array of time points and voxels]}
                    smp = {}
                    for lab in thisNim.UL: # go through each condition
                        if lab != 0: # not fixation
                            conds = np.array([l == lab for l in thisNim.labels], dtype='bool')
                            # take all time points the selected voxels
                            smp[lab] = thisNim.selectSamples(conds).samples[:,selVoxels[ROI_list[1]]]
                    
                    # collect all ROI names
                    allROIs = []
                    for thisROI in ROI_list[2]:
                        allROIs.extend(q.listDir(core.init.paths['recDir'] %subjID, \
                            pattern=thisROI+'\.nii', cropExt = True, fullPath = False))
                            
                    # prepare a list of output values that for this particular subjID and ROI
                    rec = [
                        subjID,
                        ROI_list[1], # ROI names combined together or specified by the user
                        ';'.join(allROIs), # ROI file names
                        np.sum(len(selVoxels[ROI_list[1]])) # number of voxels in the ROI
                        ]
                    
                    # finally, send all information to the classifier
                    for key, value in allConds.items():
                        # make a combination of train and test samples
                        pairs = zip(q.combinations(value[0],2), q.combinations(value[1],2))
                        # rerun classification many times to get stable results
                        for r in range(repeat):
                            print roiIter*repeat+r,
                            out = wrapperClf(smp, pairs, splits, clfList, selRange, recPre = rec + [key, roiIter*repeat+r])
                            dataOut.extend(out)
                    print ''
        
        plotData = tb.tabarray(records = dataOut, names = names)
        
        if not rp.noOutput: plotData.saveSV(rp.o)

    
    # grouping ROIs
    if rp.group:
        level = np.select([
            (plotData['ROI']=='V1') |(plotData['ROI']=='V2') |(plotData['ROI']=='V3'),
            (plotData['ROI']=='LO') |(plotData['ROI']=='pFs')],
            ['low', 'higher'])
        plotData = plotData.addcols(level, names = ['level'])
        xAxis = ['level', ['low', 'higher']]
    else: xAxis = ['ROI', [item[1] for item in rp.rois]]
    
    
    # title options    
    if rp.numVoxels == 'minRandom': title = 'random subset of voxels'
    elif rp.numVoxels == 'minCenter': title = 'voxels close the center'
    elif type(rp.numVoxels) == int: title = '%d voxels' %rp.numVoxels
    else: title = 'SVM classification performance'
    
    
    fig = da.plotTabular(
        plotData,
        xAxis = xAxis,
        categories = ['type', allConds.keys()],
        values = 'LinearNuSVMC',
        yerrName = 'subjID',
        plotType = 'bar',
        textOnly = True,
        savePlot = None,
        userOpt = {
            'colors': 'gray1',
            'hatch': ['','','/','x'],
            'subplot': {
                'title': title,
                'ylabel': 'proportion correct',
                'other': "ax.axhline(y=.5, color = '.2', ls='--', marker = 'None')",
                'xlabel': ''
                }
            }
        )
    if rp.analysis[1] != 'main': fig.set_size_inches((5,4), forward = True)
    if rp.savePlot != None: plt.savefig(rp.savePlot, bbox_inches = 'tight')
    if not rp.textOnly: plt.show()
    # da.plotTabular(
        # plotData,
        # xAxis = 'subjID',#['ROI', rp.rois],
        # categories = ['ROI', [item[1] for item in rp.rois]],
        # values = 'LinearNuSVMC',
        # yerrName = 'session',
        # plotType = 'bar',
        # textOnly = rp.textOnly,
        # savePlot = rp.savePlot,
        # filter = "plotData['type']=='whole (right)'"
        # )