###############################################################################

"""
Generic functions for any kind of an experiment
"""

###############################################################################
##                             Import modules                                ##
###############################################################################


# include PsychoPy
from psychopy import visual, core, event

# include system functions
import sys, os, getopt, re

# include the ability to read various image formats
import pygame.image
#import OpenGL.GL as gl

# include the ability to do some math
import numpy as np
import matplotlib.pyplot as plt
import tabular as tb
import random

# include ability to use comma separated values
import csv

# include timing functions
import time

# include commonly used modules and functions
import tools.qFunctions as q
import core.init as cinit


###############################################################################
##                             Define variables                              ##
###############################################################################


# fMRI parameters
TR = 2000. # msec
trigger = '5'
defaultKeys = ['escape', trigger]
defaultBuffer = 4 # Number of fixation trials at the beginning and at the end
                  # of each scan if the paradigm file does not already have
                  # them. The duration is defaultBuffer*TR
defaultFilter = "(plotData['cond'] != 0) & (plotData['accuracy'] != 'No response')"
shiftTp = 2

# Output parameters
signalDet = {False: 'Incorrect',
             True: 'Correct',
             None: 'No response'}

validResponses = {1: '1', 2: '2', 3: '3', 4: '4'} # Right button box
#validResponses = {1: '9', 2: '8', 3: '7', 4: '6'} # Left button box
keyNames = {1: 'index finger',
            2: 'middle finger',
            3: 'ring finger',
            4: 'little finger'}
             

###############################################################################
##                             Useful functions                              ##
###############################################################################


def exit(expPlan = None):
    """
    What to do when exit is requested
    """
    if expPlan != None:
        dispRunStats(expPlan)
    core.quit()
            
            
def lastKey(keyList = defaultKeys, expPlan = None):
    """
    Extract the last key pressed from the event list    
    """
    thisKeyList = event.getKeys(keyList = keyList)
    if len(thisKeyList) > 0:        
        thisKey = thisKeyList.pop()
        if thisKey == 'escape': exit(expPlan = expPlan)
        else: return thisKey
    else:
        return None

def drawLine(shapeStim):
    
    vertices = shapeStim.vertices
    if shapeStim.closeShape: numPairs = len(vertices)
    else: numPairs = len(vertices)-1
    
    shape = []
    
    wh = shapeStim.lineWidth/2
    for i in range(numPairs):
        thisPair = [shapeStim.vertices[i],shapeStim.vertices[(i+1)%len(shapeStim.vertices)]]
        edges = [thisPair[1][0]-thisPair[0][0], thisPair[1][1]-thisPair[0][1]]
        #direction = np.sign(thisPair[1][0]-thisPair[0][0])
        #parts = [np.sign(thisPair[0][0])*np.sqrt(thisPair[0][0]**2 + thisPair[0][1]**2),
        #         np.sign(thisPair[1][0])*np.sqrt(thisPair[1][0]**2 + thisPair[1][1]**2)]
        lh = np.sqrt(edges[0]**2 + edges[1]**2)/2#np.abs(parts[0])+np.abs(parts[1])# 
        
        line = visual.ShapeStim(
            win,
            lineWidth   = 0,
            lineColor   = None,
            fillColor   = shapeStim.lineColor,
            ori         = -np.arctan2(edges[1],edges[0])*180/np.pi,
            pos         = [(thisPair[0][0]+thisPair[1][0])/2, (thisPair[0][1]+thisPair[1][1])/2],
            vertices    = [[-lh,-wh],[-lh,wh],
                           [lh,wh],[lh,-wh]]
        )
        shape.append(line)
        #import pdb; pdb.set_trace()
    
    return shape
    
      
###############################################################################
##                             Runtime functions                             ##
###############################################################################


def createBasicStim(debug = False):
    """
    Generates a window from presenting stimuli. Prepares a fixation and a black
    box for easy use. They all can be accessed later on.
    """
    global win, fixation, box
    from psychopy import monitors
    
    # define monitor
    #mon = monitors.Monitor('fMRI', distance = 31.5, width = 23)
    #mon.setSizePix((1024,768))
    mon = monitors.Monitor('LEP', distance = 80, width = 37.5)    
    mon.setSizePix((1280,1024))
    
    # Define the window where everything will be presented
    win = visual.Window(
        size     = mon.getSizePix(),
        monitor  = mon,
        units    = 'deg',
        fullscr  = not debug,
        allowGUI = False, # mouse will not be seen
        pos      = [0, 0],
        viewPos  = [0, 0],
        color    = 'grey'
    )
    #import pdb; pdb.set_trace()
    # Set up fixation
    fixation = visual.PatchStim(
        win,
        color  = 'black',
        tex    = None,
        mask   = 'circle',
        size   = .3
    )

    # A box representing the size of a stimulus
    box = visual.ShapeStim(
        win,
        lineColor   = 'black',
        fillColor   = 'black',
        closeShape  = True
    )

    
def prepareExpPlan(rp, paraType, descr = 'descr', otherDescr = None,
    addBuffer = 0):
    global numTrialsInBlock

    q.listify(descr, 1)
    (paraName, paraLines) = readPara(rp.runType)
    numBlocks = len(paraLines)
    out = []
    if addBuffer > 0: extension = [0] * addBuffer
    else: extension = []
    
    # Unpack blocks into individual trials
    out.extend(extension)
    for thisLine in paraLines:
        out.extend([thisLine] * numTrialsInBlock)
    out.extend(extension)
    
    numTrials = len(out)
    # bufferTrue = []
    # for n in range(numTrials):
        # blockNo = 
        # if n in range(addBuffer) or n in range(numTrials - addBuffer, numTrials)
            # blockNo = 
        # else:
            # bufferTrue.append(False)
    # Deal with addBuffer dependent outputs    
    blockNo = range(0, numBlocks) * numTrialsInBlock    
    blockNo.sort()
    startBlock = [1] + [0]*(numTrialsInBlock-1)
    startBlock = startBlock*numBlocks
    if addBuffer > 0:
        if paraType == 'blocked':
            blockNo = [0]*addBuffer + (np.array(blockNo)+1).tolist() + \
                [numBlocks]*addBuffer
            startBlock = [1]+[0]*(addBuffer-1) + startBlock + \
                [1]+[0]*(addBuffer-1)
        elif paraType == 'event':
            blockNo = range(0, len(blockNo) + 2*addBuffer)
            startBlock = [1]*addBuffer + startBlock + [1]*addBuffer
        
    durs = [TR/1000.]*addBuffer + [trialDur]*(numTrials-2*addBuffer) + \
        [TR/1000.]*addBuffer
    onsets = np.cumsum([0]+durs[:-1])
    actualOnsets = np.empty((numTrials,), '<f8')
        
    firstFew = [rp.expName, rp.subjID, \
        rp.runNo, rp.runType, paraType, paraName]
    firstArray = np.array(firstFew*numTrials)
    firstArray = firstArray.reshape(numTrials, len(firstFew))

    expPlan = tb.tabarray(array = firstArray,
        names = ['expName','subjID','sessNo','runType','paraType','paraName'])
    expPlan = expPlan.addcols(
        [blockNo, startBlock, range(numTrials), out, onsets, actualOnsets, durs],
        names = ['blockNo','startBlock','trialNo','cond','onset','actualOnset','dur'])
    
    # Create new colums with condition descriptors
    for i, itemName in enumerate(q.listify(descr)):
        newList = [paraTable[cond][i] for cond in expPlan['cond']]
        for b in range(addBuffer):
            newList[b] = 'Buffer'
            newList[-b-1] = 'Buffer'

        expPlan = expPlan.addcols(newList, names = itemName)

    expPlan = expPlan.addcols(
        np.empty((numTrials,), '|O4'), names = ['corrResp'])
    expPlan = expPlan.addcols(
        np.empty((numTrials,), '|O4'), names = ['subjResp'])
    expPlan = expPlan.addcols(
        np.empty((numTrials,), '|S12'), names = ['accuracy'])
    expPlan = expPlan.addcols(
        np.empty((numTrials,), '|O4'), names = ['RT'])
    if otherDescr != None:
        for (itemName, itemType) in q.listify(otherDescr,1):
            expPlan = expPlan.addcols(
                np.empty((numTrials,), itemType), names = [itemName])    
    expPlan = expPlan.addcols(np.empty((numTrials,), '|O4'),
        names = ['trialStruct'])

    return expPlan
    

def makeTrial(expPlan, trial, s, p, assignFun = None):
    # Assign values to trialStruct
    dataNames = list(expPlan.dtype.names)
    descr = dataNames[dataNames.index('dur') + 1]
    #defAct = [s[0], ["setPos(" + str(p[0]) + ")"]]
    if s.get(0) == None: s[0] = fixation
    if p.get(0) == None: p[0] = list(fixation.pos)
    
    for lineNo, thisLine in enumerate(expPlan):   
        key = thisLine['cond']
        trialStruct = []
        if thisLine[descr] == 'Buffer' or key == 0:
            eventStruct = {}
            eventStruct['dur'] = thisLine['dur']
            eventStruct['display'] = [[s[0], ["setPos(" + str(p[0]) + ")"]]]
            trialStruct.append(eventStruct)
        else:
            for thisEvent in trial:
                eventStruct = {}
                eventStruct['dur'] = float(thisEvent[0])/1000
                
                # go through each position in the position list (i.e., display)
                # and couple each position with the correct stimulus
                temp = []
                for ind, pos in enumerate(thisEvent[1]):                    
                    if pos == 0 or assignFun == None:
                        temp.append([s[0], ["setPos(" + str(p[0]) + ")"]])
                    else: temp.append(assignFun(lineNo, thisLine, ind, pos))
                    
                eventStruct['display'] = temp
                trialStruct.append(eventStruct)
            
        expPlan[lineNo]['trialStruct'] = trialStruct
    return expPlan

    
def readPara(runType):
    """
    Reads in a paradigm file
    """
    
    # Pick a paradigm file
    paraUsed = cinit.getParaList(runType)
    paraAll = q.listDir(cinit.paths['paraDirSup'], ending = '.txt', \
        cropExt = True, fullPath = False)
    # if all paras have been used already, take the lastest ones only
    #if len(paraUsed) >= len(paraAll):
    paraUsed = paraUsed[len(paraUsed) - len(paraUsed) % len(paraAll):]
    paraRemain = list(set(paraAll).difference(set(paraUsed)))
    if paraRemain == []:
        paraRemain = paraAll
    
    # Declare para reader
    paraName = random.choice(paraRemain)
    paraFile = open(cinit.paths['paraDirSup'] + paraName + '.txt')
    paraLines = paraFile.read().split('\n')
    paraLines = [int(cond) for cond in paraLines]
    paraFile.close
        
    return (paraName, paraLines)
    

def makePara(expPlan):
    """
    Makes a .m file for generating the SPM para file
    """
    global numTrialsInBlock, trialDur, TR
    def sortJoin(x):
        global TR
        x.sort()
        s = [str(int(round(item * 1000. / TR))) for item in x]
        return ' '.join(s)
    
    block = np.ones(expPlan.shape[0], dtype = np.int)    
    if expPlan[0]['paraType'] == 'blocked':        
        SPMblockLen = str(int(numTrialsInBlock * trialDur * 1000. / TR))
        # lastCond = None
        # for i, cond in enumerate(expPlan['cond']):
            # if cond != lastCond:
                # block[i] = 1
                # lastCond = cond
            # else:
                # block[i] = 0        
    else:        
        SPMblockLen = '0'
        
    #allData = expPlan.addcols(block, names=['block'])
    dataNames = list(expPlan.dtype.names)
    descr = dataNames[dataNames.index('dur') + 1:dataNames.index('corrResp')]
    #descr = dataNames[dataNames.index('dur') + 1]
    allData = expPlan[['startBlock', 'cond', 'onset'] + descr]\
        [expPlan[descr[0]] != 'Buffer']
    
    agg = allData[allData['startBlock']==1].aggregate(
        On = 'cond', AggFunc = lambda x: x[0],
        AggFuncDict = {'onset': lambda x: sortJoin(x)})
    aggExtractOnset = agg[['onset']].extract()
    aggExtractDescr = agg[q.listify(descr)].extract()
    
    out = {'names': [], 'onsets': [], 'durations': []}    
    for onset, names in zip(aggExtractOnset, aggExtractDescr):
        out['durations'].append(SPMblockLen)
        out['onsets'].append("[" + onset + "]")
        out['names'].append(' | '.join(q.listify(names)))
    
    lenName = 0
    for name in out['names']:
        if len(name) > lenName:
            lenName = len(name)
    out['names'] = ["'" + it.ljust(lenName) + "'" for it in out['names']]
    
    if not os.path.isdir(cinit.paths['paraDir']):
        os.makedirs(cinit.paths['paraDir'])
    fileName = '_'.join(['para','%02d' %(int(expPlan[0]['sessNo'])),
        expPlan[0]['runType']])
    paraFile = open(cinit.paths['paraDir'] + '/' + fileName + '.m', 'w')
    for (key, item) in out.items():
        paraFile.write(key + " = {" + ", ".join(item) + "};\n")
    paraFile.write("save('" + fileName + '.mat' +
        "', 'names', 'onsets', 'durations');")
    paraFile.close()
    
    
def showIntro(text = None, args = None):
    """
    Show instructions until trigger is received
    """
    global win, box, keyNames, trigger
    
    if args != None:
        (vertices, p, stimSize) = args
        # add response keys and stimuli-size boxes at the stimuli position
        box.setVertices(vertices)
        for (key, value) in keyNames.items():
            box.setPos(p[key])
            box.draw()
            number = visual.TextStim(win, text = value,
                                     pos = p[key], wrapWidth = stimSize - 1)
            number.draw()

    if text != None:
        introText = visual.TextStim(
            win,
            text = text,
            color = (-1,-1,-1),
            height = 1,
            wrapWidth = 10)
        introText.draw()
        
    win.flip()
    thisKey = None
    while thisKey != trigger:
        thisKey = lastKey()    

        
def waitEvent(globClock, trialClock, eventClock,
    expPlan, thisTrial, thisEvent, thisDisplay):
    """
    Default waiting function for the event
    Does nothing but catching key input of default keys (escape and trigger)
    """
    while eventClock.getTime() < thisEvent['dur'] and \
    trialClock.getTime() < thisTrial['dur'] and \
    globClock.getTime() < thisTrial['onset'] + thisTrial['dur']:
        
        lastKey(expPlan = expPlan)
    
            
def run(expPlan, noOutput = False, defaultFun = waitEvent):
    """
    Iterate over the sequence of events
    """
    global win, validResponses, signalDet, paths
    
    outputNames = []
    outputNames.extend(expPlan.dtype.names[:-1]) # all but trialStruct    
    if not noOutput:
        makePara(expPlan)
        if not os.path.isdir(cinit.paths['dataBehavDir']):
            os.makedirs(cinit.paths['dataBehavDir'])
        fileName = '_'.join(['data','%02d' %int(expPlan[0]['sessNo']), expPlan[0]['runType']])
        dataFile = open(cinit.paths['dataBehavDir'] + '/' + fileName + '.csv', 'w')
        dataCSV = csv.writer(dataFile, lineterminator = '\n')
        dataCSV.writerow(expPlan.dtype.names[:-1])
        
    globClock = core.Clock()
    trialClock = core.Clock()
    eventClock = core.Clock()
    for thisTrial in expPlan:
        trialClock.reset()
        thisResp = [None, None]
        
        for j, thisEvent in enumerate(thisTrial['trialStruct']):
            eventClock.reset()
            thisDisplay = drawDisplay(thisEvent['display'])
            win.flip()
            if j == 0: thisTrial['actualOnset'] = globClock.getTime()
            
            defaultFun(globClock, trialClock, eventClock,
                expPlan[:thisTrial['trialNo']+1], thisTrial, thisEvent,
                thisDisplay)
            
            allKeys = event.getKeys(
                keyList = validResponses.values(),
                timeStamped = trialClock)
            if len(allKeys) > 0: thisResp = allKeys.pop()
            clearDisplay(thisDisplay, thisEvent['display'])
            
        thisTrial['subjResp'] = thisResp[0]
        thisTrial['accuracy'] = signalDet[q.advCmp(thisTrial['corrResp'],
                                                 thisTrial['subjResp'])]
        if thisResp[1] != None: thisTrial['RT'] = thisResp[1]*1000 #convert to msec
        
        if not noOutput: dataCSV.writerow(list(thisTrial)[:-1])
    
    if not noOutput: dataFile.close()
    win.close()
    
    return expPlan
    
    
#def drawDisplay(display):
#    """
#    Draws stimuli on the screen
#    Does not update the screen!
#    Take each component of the stimuli and make put them in one big list
#    that makes them all independent, so that we can freely specify the pos
#    """    
#    itemList = []
#    for items in display:
#        temp = []
#        #
#        for item in items[1]:
#            if type(item) in [tuple, list]:
#                temp.append(item[0] + '(' + str(item[1]) + ')')
#                
#            else:
#                temp.append(item)                

#        for stim in items[0]:
#            for op in temp:
#                eval('stim.' + op)
#            stim.draw()
#            itemList.append(stim)
#            
#    return itemList


def drawDisplay(display):
    """
    Draws stimuli on the screen

    Does not update the screen!
    Take each component of the stimuli and make put them in one big list
    that makes them all independent, so that we can freely specify the pos

    """    
    itemList = []
    for items in display:
        for stim in items[0]:        
            for item in items[1]:
                if type(item) in [tuple, list]:
                    #import pdb; pdb.set_trace()
                    if item[0] == 'setPos':
                        eval('stim.' + item[0] + '(' + str([item[1][0]+stim.pos[0], item[1][1]+stim.pos[1]]) + ')')
                    else: eval('stim.' + stim + '(' + str(item[1]) + ')')
                    
                else:
                    eval('stim.' + item)
                    
            stim.draw()
            if item[0] == 'setPos':
                eval('stim.' + item[0] + '(' + str([-item[1][0]+stim.pos[0], -item[1][1]+stim.pos[1]]) + ')')
            itemList.append(stim)
            
    return itemList
    
    
# def clearDisplay(itemList):
    # """
    # Clears textures of given stimuli in order to free memory
    # """    
    # for item in itemList:
        # try:
            # item[0].clearTextures
        # except:
            # pass


def clearDisplay(itemList, display):
    """
    Clears textures of given stimuli in order to free memory
    """    
    # for items in display:
        # for stim in items[0]:
            # for item in items[1]:
                # if item[0] == 'setPos':
                    # eval('stim.' + item[0] + '(' + str([-item[1][0]+stim.pos[0], -item[1][1]+stim.pos[1]]) + ')')
                
    for item in itemList:
        try:
            item[0].clearTextures
        except:
            pass
             
    
# def pacemaker(thisKey, blockLen, trigger = trigger, triggerCount = 0):
    # """
    # Advance to the next trial
    # when the blockLen number of triggers is accumulated
    # """
    # if thisKey == trigger:
        # triggerCount += 1
    # if triggerCount >= blockLen:
        # triggerCount = 0
        # return True
    # else:
        # return False
            
            
# def trialData(runParams, paraName, thisTrial, trialNo, TR = TR, signalDet = signalDet):
    # """
    # Combine all trial data into one list for easy output
    # """
    # thisTrialData = []
    # thisTrialData.extend(runParams[0:3])
    # thisTrialData.extend([paraName, trialNo, trialNo*TR/1000.0, thisTrial['para']])
    # thisTrialData.extend(thisTrial['description'])
    # thisTrialData.extend([
        # thisTrial['corrResp'],
        # thisTrial['respKey'],                
        # signalDet[advCmp(thisTrial['corrResp'], thisTrial['respKey'])],
        # thisTrial['respTime']
    # ])
    # return thisTrialData

    
def dispRunStats(expPlan):
    """
    Prints out RT and percent correct response
    """
    expPlan = expPlan.deletecols(['trialStruct'])
    onsetVar = expPlan[['actualOnset']].extract() - expPlan[['onset']].extract()
    
    print
    print '==== Run Stats ===='
    print 'Onset diff: max = %.4f and min = %.4f' %(onsetVar.max(), onsetVar.min())
    print 'No response: %d' %expPlan[expPlan['accuracy'] == 'No response'].shape[0]
