###############################################################################

"""
Presents 4 shapes in four quadrants
Task: Indicate which one of the four shapes is different
"""

# Get default experimental flow variables and functions
import sys, os
import core
import core.exp as exp


def runExp(rp):
    
    # Include PsychoPy
    from psychopy import visual    
    
    ## Variables ##############################################################
    paraType = 'event'
    # fMRI parameters
    exp.numTrialsInBlock = 1

    #             CondNo Context     Pos
    exp.paraTable = {0: ('Fixation', ''           ),
                     1: ('None',    'Top Left'    ),
                     2: ('None',    'Top Right'   ),
                     3: ('None',    'Bottom Left' ),
                     4: ('None',    'Bottom Right'),
                     5: ('Good',    'Top Left'    ),
                     6: ('Good',    'Top Right'   ),
                     7: ('Good',    'Bottom Left' ),
                     8: ('Good',    'Bottom Right')}

    # Stimuli parameters
    stimSize = 4
    stimWidth = .3 # px; the weight of the line
    stimDist = 5 # from the fixation in x or y dir
    stimColor = 'black'
    
    stimSizeHalf = stimSize/2
    stimWhalf = stimWidth/2
    
    # Define stimuli
    exp.createBasicStim(rp.debug)
       
    diag45 = exp.drawLine(visual.ShapeStim(
        exp.win,
        lineColor   = stimColor,
        lineWidth   = stimWidth,
        fillColor   = stimColor,
        closeShape  = False,
        vertices    = [[-stimSizeHalf, -stimSizeHalf],
                      [ stimSizeHalf,  stimSizeHalf]]
    ))
    diag135 = exp.drawLine(visual.ShapeStim(
        exp.win,
        lineColor   = stimColor,
        lineWidth   = stimWidth,
        fillColor   = stimColor,
        vertices    = [[-stimSizeHalf, stimSizeHalf],
                      [ stimSizeHalf,-stimSizeHalf]]
    ))
    corner = exp.drawLine(visual.ShapeStim(
        exp.win,
        lineColor   = stimColor,
        lineWidth   = stimWidth,
        fillColor   = None,
        vertices    = [[-stimSizeHalf, stimSizeHalf],
                       [-stimSizeHalf, -stimSizeHalf],
                       [ stimSizeHalf, -stimSizeHalf]],
        closeShape  = False
    ))
    
    # All stimuli
    s = {0: [exp.fixation],
         1: diag45,
         2: diag135,
         3: diag45 + corner,
         4: diag135 + corner}

    # Positions of all stimuli
    p = {0: (        0,        0), # Fixation position
         1: (-stimDist, stimDist),
         2: ( stimDist, stimDist),
         3: (-stimDist,-stimDist),
         4: ( stimDist,-stimDist)}
        
    events = {1: ( 150, [1, 2, 3, 4, 0]),
              0: (1850, [0            ])}
    trial = [events[1], events[0]]
    exp.trialDur = sum(event[0] for event in events.values())/1000.
    
    
    ## Functions ##############################################################
    def assignFun(lineNo, thisLine, ind, pos):
        key = thisLine['cond']
        ind2 = ((key-1)/4)*2+(ind==(key-1)%4)+1
        return [s[ind2], [['setPos',p[pos]]] ] #[["setPos("+str(p[pos])+")"]]]
    

    ## Main flow ##############################################################
        
    # Generate a list of image pairs and display positions
    expPlan = exp.prepareExpPlan(rp, paraType,
        descr = ['context', 'pos'], addBuffer = 4)
    
    # Assign values to context, pos, and corrResp
    for (key, value) in exp.paraTable.items():
        if key == 0:
            expPlan['corrResp'][expPlan['cond'] == key] = \
                exp.validResponses.get(0)
        else:
            expPlan['corrResp'][expPlan['cond'] == key] = \
                exp.validResponses.get((key - 1) % 4 + 1)
    
    # Create each trial
    expPlan = exp.makeTrial(expPlan, trial, s, p, assignFun = assignFun)
 
    # Make intro screen and start running
    text = "task: indicate which one of the four shapes is different"
    vertices = [[-stimSizeHalf, -stimSizeHalf],
                [-stimSizeHalf,  stimSizeHalf],
                [ stimSizeHalf,  stimSizeHalf],
                [ stimSizeHalf, -stimSizeHalf]]    
    exp.showIntro(text = text, args = (vertices, p, stimSize))
    expPlan = exp.run(expPlan, rp.noOutput)
    plot(rp)
    exp.dispRunStats(expPlan)
            

def plot(rp, plotData = None):    
    """
    Plots accuracy and response time
    
    **Parameters**
    
        rp: Namespace (required)
            Run parameters that are parsed from the command line
        plotData: tabarray
            Data that is to be plotted
        
    """
    
    import analysis.dataAnalysis
    import analysis.behav
    
    if len(rp.analysis) > 1: plotWhat = rp.analysis[1]
    else: plotWhat = 'mean'
    
    # get data and store it in plotData
    if plotData == None: plotData = analysis.behav.getBehavData(rp)
    
    if plotWhat == 'mean':
    
        analysis.dataAnalysis.plotTabular(
            plotData,
            filter = exp.defaultFilter,
            xAxis = 'expName',
            categories = ['context', [('None','parts'), ('Good','whole')] ],
            values = 'accuracy',
            valueFilter = "plotData['accuracy'] == 'Correct'",
            yerrName = 'subjID',
            plotType = 'bar',
            textOnly = rp.textOnly,
            savePlot = rp.savePlot,
            userOpt = {
                'colors': 'gray1',
                'subplot': {
                    'title': 'behavioral performance',
                    'ylabel': 'proportion correct'
                    }
                }
        )
    
        analysis.dataAnalysis.plotTabular(
            plotData,
            filter = exp.defaultFilter + "& (plotData['accuracy'] == 'Correct')",
            xAxis = 'expName',
            categories = ['context', [('None','parts'), ('Good','whole')] ],
            values = 'RT',
            yerrName = 'subjID',
            plotType = 'bar',
            textOnly = rp.textOnly,
            savePlot = rp.savePlot,
            userOpt = {
                'colors': 'gray1',
                'subplot': {
                    'ylabel': 'ms',
                    'title': 'response time'
                    }
                }
        )
    
    elif plotWhat == 'byRun':
        analysis.dataAnalysis.plotTabular(
            plotData,
            filter = exp.defaultFilter,
            xAxis = 'runNo',
            categories = ['context', [('None','parts'), ('Good','whole')] ],
            values = 'accuracy',
            valueFilter = "plotData['accuracy'] == 'Correct'",
            yerrName = 'subjID',
            plotType = 'bar',
            textOnly = rp.textOnly,
            savePlot = rp.savePlot,
            userOpt = {
                'colors': 'gray1',
                'subplot': {
                    'xlabel': 'testing runs',
                    'ylabel': 'proportion correct'
                    }
                }
        )    
    
        analysis.dataAnalysis.plotTabular(
            plotData,
            filter = exp.defaultFilter + "& (plotData['accuracy'] == 'Correct')",
            xAxis = 'runNo',
            categories = ['context', [('None','parts'), ('Good','whole')] ],
            values = 'RT',
            yerrName ='subjID',
            plotType = 'bar',
            textOnly = rp.textOnly,
            savePlot = rp.savePlot,
            userOpt = {
                'colors': 'gray1',
                'subplot': {
                    'xlabel': 'testing runs',
                    'ylabel': 'ms',
                    'title': 'response time'
                    }
                }
        )
    
         
def run(rp):
    runExp(rp)